@@ -1,6 +1,9 @@
 package config
 
 import (
+	"os"
+	"os/exec"
+	"path/filepath"
 	"testing"
 
 	"github.com/stretchr/testify/assert"
@@ -10,16 +13,15 @@ import (
 func TestAddIgnitionFile_MultipleOptions(t *testing.T) {
 	vm := &VirtualMachine{}
 	err := vm.AddIgnitionFileFromCmdLine("file1,file2")
-	assert.EqualError(t, err, "ignition only accept one option in command line argument")
+	assert.EqualError(t, err, "ignition only accepts one option in command line argument")
 }
 
 func TestAddIgnitionFile_OneOption(t *testing.T) {
 	vm := &VirtualMachine{}
 	err := vm.AddIgnitionFileFromCmdLine("file1")
 	require.NoError(t, err)
-	assert.Len(t, vm.Devices, 1)
-	assert.Equal(t, uint32(ignitionVsockPort), vm.Devices[0].(*VirtioVsock).Port)
 	assert.Equal(t, "file1", vm.Ignition.ConfigPath)
+	assert.Equal(t, ignitionVsockPort, vm.Ignition.VsockPort)
 }
 
 func TestNetworkBlockDevice(t *testing.T) {
@@ -44,3 +46,26 @@ func TestNetworkBlockDevice_NoDevice(t *testing.T) {
 	nbdItem := vm.NetworkBlockDevice("nbd2")
 	require.Nil(t, nbdItem)
 }
+
+func TestVirtualMachine_ValidateBlockDevices(t *testing.T) {
+	vm := &VirtualMachine{}
+
+	tmpDir := t.TempDir()
+	imagePath := filepath.Join(tmpDir, "disk.qcow2")
+	size := "1G"
+
+	cmd := exec.Command("qemu-img", "create", "-f", "qcow2", imagePath, size)
+	err := cmd.Run()
+
+	require.NoError(t, err)
+	defer os.Remove(imagePath)
+
+	dev, err := VirtioBlkNew(imagePath)
+	require.NoError(t, err)
+	vm.Devices = append(vm.Devices, dev)
+
+	err = dev.validate()
+	require.Error(t, err)
+
+	require.ErrorContains(t, err, "vfkit does not support qcow2 image format")
+}
