@@ -4,21 +4,12 @@ import (
 	"encoding/json"
 	"fmt"
 	"reflect"
+	"slices"
 	"testing"
 
 	"github.com/stretchr/testify/require"
 )
 
-func contains(strings []string, val string) bool {
-	for _, str := range strings {
-		if val == str {
-			return true
-		}
-	}
-
-	return false
-}
-
 // This sets all the fields of the `obj` struct to non-empty values.
 // This will be used to test JSON serialization as extensively as possible to
 // avoid breaking backwards compatibility.
@@ -34,7 +25,7 @@ func fillStruct(t *testing.T, obj interface{}, skipFields []string) {
 		fieldVal := val.FieldByIndex(e.Index)
 		typeName := val.Type().Name()
 
-		if contains(skipFields, field.Name) {
+		if slices.Contains(skipFields, field.Name) {
 			continue
 		}
 		switch fieldVal.Kind() {
@@ -91,9 +82,10 @@ var jsonTests = map[string]jsonTest{
 			ignition, err := IgnitionNew("config", "socket")
 			require.NoError(t, err)
 			vm.Ignition = ignition
+			ignition.VsockPort = 0 // slight hack for the test to pass as VsockPort is not serialized
 			return vm
 		},
-		expectedJSON: `{"vcpus":3,"memoryBytes":4194304000,"bootloader":{"kind":"linuxBootloader","vmlinuzPath":"/vmlinuz","initrdPath":"/initrd","kernelCmdLine":"console=hvc0"}, "ignition":{"kind":"ignition","configPath":"config","socketPath":"socket"}}`,
+		expectedJSON: `{"vcpus":3,"memoryBytes":4194304000,"bootloader":{"kind":"linuxBootloader","vmlinuzPath":"/vmlinuz","initrdPath":"/initrd","kernelCmdLine":"console=hvc0"}, "ignition":{"kind":"ignition","configPath":"config"}}`,
 	},
 	"TestVirtioRNG": {
 		newVM: func(t *testing.T) *VirtualMachine {
@@ -231,7 +223,7 @@ var jsonStabilityTests = map[string]jsonStabilityTest{
 
 			return vm
 		},
-		skipFields:   []string{"Bootloader", "Devices", "Timesync", "Ignition"},
+		skipFields:   []string{"Bootloader", "Devices", "Timesync", "Ignition", "Nested", "PidFile"},
 		expectedJSON: `{"vcpus":3,"memoryBytes":3,"bootloader":{"kind":"linuxBootloader","vmlinuzPath":"/vmlinuz","kernelCmdLine":"console=hvc0","initrdPath":"/initrd"},"devices":[{"kind":"virtiorng"}],"timesync":{"vsockPort":1234}}`,
 	},
 	"RosettaShare": {
@@ -249,7 +241,7 @@ var jsonStabilityTests = map[string]jsonStabilityTest{
 	"VirtioNet": {
 		obj:          &VirtioNet{},
 		skipFields:   []string{"Socket"},
-		expectedJSON: `{"kind":"virtionet","nat":true,"unixSocketPath":"UnixSocketPath","macAddress":"00:11:22:33:44:55"}`,
+		expectedJSON: `{"kind":"virtionet","nat":true,"unixSocketPath":"UnixSocketPath","vfkitMagic":true,"macAddress":"00:11:22:33:44:55"}`,
 	},
 	"VirtioRNG": {
 		obj:          &VirtioRng{},
@@ -257,7 +249,7 @@ var jsonStabilityTests = map[string]jsonStabilityTest{
 	},
 	"VirtioSerial": {
 		obj:          &VirtioSerial{},
-		expectedJSON: `{"kind":"virtioserial","logFile":"LogFile","usesStdio":true}`,
+		expectedJSON: `{"kind":"virtioserial","logFile":"LogFile","ptyName":"PtyName","usesPty":true,"usesStdio":true}`,
 	},
 	"VirtioVsock": {
 		obj:          &VirtioVsock{},
@@ -285,29 +277,32 @@ var jsonStabilityTests = map[string]jsonStabilityTest{
 		newObjectFunc: func(t *testing.T) any {
 			blk, err := VirtioBlkNew("")
 			require.NoError(t, err)
+			blk.Type = DiskBackendImage
 			return blk
 		},
 
-		skipFields:   []string{"DevName", "URI"},
-		expectedJSON: `{"kind":"virtioblk","devName":"virtio-blk","imagePath":"ImagePath","readOnly":true,"deviceIdentifier":"DeviceIdentifier"}`,
+		skipFields:   []string{"DevName", "URI", "Type"},
+		expectedJSON: `{"kind":"virtioblk","devName":"virtio-blk","imagePath":"ImagePath","readOnly":true,"type":"image","deviceIdentifier":"DeviceIdentifier"}`,
 	},
 	"USBMassStorage": {
 		newObjectFunc: func(t *testing.T) any {
 			usb, err := USBMassStorageNew("")
 			require.NoError(t, err)
+			usb.Type = DiskBackendImage
 			return usb
 		},
-		skipFields:   []string{"DevName", "URI"},
-		expectedJSON: `{"kind":"usbmassstorage","devName":"usb-mass-storage","imagePath":"ImagePath","readOnly":true}`,
+		skipFields:   []string{"DevName", "URI", "Type"},
+		expectedJSON: `{"kind":"usbmassstorage","devName":"usb-mass-storage","imagePath":"ImagePath","readOnly":true,"type":"image"}`,
 	},
 	"NVMExpressController": {
 		newObjectFunc: func(t *testing.T) any {
 			nvme, err := NVMExpressControllerNew("")
 			require.NoError(t, err)
+			nvme.Type = DiskBackendImage
 			return nvme
 		},
-		skipFields:   []string{"DevName", "URI"},
-		expectedJSON: `{"kind":"nvme","devName":"nvme","imagePath":"ImagePath","readOnly":true}`,
+		skipFields:   []string{"DevName", "URI", "Type"},
+		expectedJSON: `{"kind":"nvme","devName":"nvme","imagePath":"ImagePath","readOnly":true,"type":"image"}`,
 	},
 	"LinuxBootloader": {
 		obj:          &LinuxBootloader{},
@@ -339,6 +334,55 @@ type jsonStabilityTest struct {
 	expectedJSON  string
 }
 
+func testVirtioNetVfkitMagicJson(t *testing.T, vfkitMagic bool, useVfkitMagicDefault bool) {
+	const (
+		jsonVfkitMagicDefault = `{"vcpus":3,"memoryBytes":4194304000,"bootloader":{"kind":"linuxBootloader","vmlinuzPath":"/vmlinuz","initrdPath":"/initrd","kernelCmdLine":"console=hvc0"},"devices":[{"kind":"virtionet","nat":false,"unixSocketPath":"/some/path/to/socket","macAddress":"00:11:22:33:44:55"}]}`
+		jsonVfkitMagicFalse   = `{"vcpus":3,"memoryBytes":4194304000,"bootloader":{"kind":"linuxBootloader","vmlinuzPath":"/vmlinuz","initrdPath":"/initrd","kernelCmdLine":"console=hvc0"},"devices":[{"kind":"virtionet","nat":false,"unixSocketPath":"/some/path/to/socket","macAddress":"00:11:22:33:44:55","vfkitMagic":false}]}`
+		jsonVfkitMagicTrue    = `{"vcpus":3,"memoryBytes":4194304000,"bootloader":{"kind":"linuxBootloader","vmlinuzPath":"/vmlinuz","initrdPath":"/initrd","kernelCmdLine":"console=hvc0"},"devices":[{"kind":"virtionet","nat":false,"unixSocketPath":"/some/path/to/socket","macAddress":"00:11:22:33:44:55","vfkitMagic":true}]}`
+	)
+	var jsonStr string
+	var unmarshalledVM VirtualMachine
+
+	if useVfkitMagicDefault {
+		jsonStr = jsonVfkitMagicDefault
+	} else {
+		if vfkitMagic {
+			jsonStr = jsonVfkitMagicTrue
+		} else {
+			jsonStr = jsonVfkitMagicFalse
+		}
+	}
+	err := json.Unmarshal([]byte(jsonStr), &unmarshalledVM)
+	require.NoError(t, err)
+	netDevs := unmarshalledVM.VirtioNetDevices()
+	require.Len(t, netDevs, 1)
+	require.Equal(t, vfkitMagic, netDevs[0].VfkitMagic)
+
+	vm := newLinuxVM(t)
+	dev, err := VirtioNetNew("00:11:22:33:44:55")
+	require.NoError(t, err)
+	dev.SetUnixSocketPath("/some/path/to/socket")
+	if !useVfkitMagicDefault {
+		dev.VfkitMagic = vfkitMagic
+	}
+	err = vm.AddDevice(dev)
+	require.NoError(t, err)
+	require.Equal(t, vfkitMagic, dev.VfkitMagic)
+
+	require.Equal(t, *vm, unmarshalledVM)
+}
+
+func TestVirtioNetBackwardsCompat(t *testing.T) {
+	/* Check that the vfkitMagic default is true when deserializing json */
+	t.Run("VfkitMagicJsonDefault", func(t *testing.T) { testVirtioNetVfkitMagicJson(t, true, true) })
+
+	/* Check that the vfkitMagic default can be overridden */
+	t.Run("VfkitMagicJsonDefaultOverride", func(t *testing.T) { testVirtioNetVfkitMagicJson(t, false, false) })
+
+	/* Check that explicitly setting vfkitMagic to true works as expected */
+	t.Run("VfkitMagicJsonExplicitDefault", func(t *testing.T) { testVirtioNetVfkitMagicJson(t, true, false) })
+}
+
 func TestJSON(t *testing.T) {
 	t.Run("json", func(t *testing.T) {
 		for name := range jsonTests {
