@@ -6,13 +6,14 @@ import (
 	"os"
 	"path/filepath"
 	"strings"
-	"syscall"
 
 	"github.com/crc-org/vfkit/pkg/config"
+	"github.com/crc-org/vfkit/pkg/util"
+	"golang.org/x/sys/unix"
 
 	"github.com/Code-Hex/vz/v3"
+	"github.com/pkg/term/termios"
 	log "github.com/sirupsen/logrus"
-	"golang.org/x/sys/unix"
 )
 
 // vf will define toVZ() and AddToVirtualMachineConfig() methods on these types
@@ -36,7 +37,7 @@ type vzNetworkBlockDevice struct {
 }
 
 func (dev *NVMExpressController) toVz() (vz.StorageDeviceConfiguration, error) {
-	var storageConfig DiskStorageConfig = DiskStorageConfig(dev.DiskStorageConfig)
+	var storageConfig = DiskStorageConfig(dev.DiskStorageConfig)
 	attachment, err := storageConfig.toVz()
 	if err != nil {
 		return nil, err
@@ -61,7 +62,7 @@ func (dev *NVMExpressController) AddToVirtualMachineConfig(vmConfig *VirtualMach
 }
 
 func (dev *VirtioBlk) toVz() (vz.StorageDeviceConfiguration, error) {
-	var storageConfig DiskStorageConfig = DiskStorageConfig(dev.DiskStorageConfig)
+	var storageConfig = DiskStorageConfig(dev.DiskStorageConfig)
 	attachment, err := storageConfig.toVz()
 	if err != nil {
 		return nil, err
@@ -86,7 +87,16 @@ func (dev *VirtioBlk) AddToVirtualMachineConfig(vmConfig *VirtualMachineConfigur
 	if err != nil {
 		return err
 	}
-	log.Infof("Adding virtio-blk device (imagePath: %s)", dev.ImagePath)
+	var diskType string
+	if !dev.Type.IsValid() {
+		return fmt.Errorf("invalid disk type: %s", dev.Type)
+	}
+	if dev.Type == config.DiskBackendDefault {
+		diskType = string(config.DiskBackendImage)
+	} else {
+		diskType = string(dev.Type)
+	}
+	log.Infof("Adding virtio-blk device (imagePath: %s, type: %s)", dev.ImagePath, diskType)
 	vmConfig.storageDevicesConfiguration = append(vmConfig.storageDevicesConfiguration, storageDeviceConfig)
 
 	return nil
@@ -246,39 +256,63 @@ func unixFd(fd uintptr) int {
 // https://developer.apple.com/documentation/virtualization/running_linux_in_a_virtual_machine#3880009
 func setRawMode(f *os.File) error {
 	// Get settings for terminal
-	attr, _ := unix.IoctlGetTermios(unixFd(f.Fd()), unix.TIOCGETA)
+	var attr unix.Termios
+	if err := termios.Tcgetattr(f.Fd(), &attr); err != nil {
+		return err
+	}
 
 	// Put stdin into raw mode, disabling local echo, input canonicalization,
 	// and CR-NL mapping.
-	attr.Iflag &^= syscall.ICRNL
-	attr.Lflag &^= syscall.ICANON | syscall.ECHO
-
-	// Set minimum characters when reading = 1 char
-	attr.Cc[syscall.VMIN] = 1
-
-	// set timeout when reading as non-canonical mode
-	attr.Cc[syscall.VTIME] = 0
+	attr.Iflag &^= unix.ICRNL
+	attr.Lflag &^= unix.ICANON | unix.ECHO
 
 	// reflects the changed settings
-	return unix.IoctlSetTermios(unixFd(f.Fd()), unix.TIOCSETA, attr)
+	return termios.Tcsetattr(f.Fd(), termios.TCSANOW, &attr)
 }
 
 func (dev *VirtioSerial) toVz() (*vz.VirtioConsoleDeviceSerialPortConfiguration, error) {
 	var serialPortAttachment vz.SerialPortAttachment
-	var err error
-	if dev.UsesStdio {
+	var retErr error
+	switch {
+	case dev.UsesStdio:
 		if err := setRawMode(os.Stdin); err != nil {
 			return nil, err
 		}
-		serialPortAttachment, err = vz.NewFileHandleSerialPortAttachment(os.Stdin, os.Stdout)
-	} else {
-		serialPortAttachment, err = vz.NewFileSerialPortAttachment(dev.LogFile, false)
+		serialPortAttachment, retErr = vz.NewFileHandleSerialPortAttachment(os.Stdin, os.Stdout)
+	default:
+		serialPortAttachment, retErr = vz.NewFileSerialPortAttachment(dev.LogFile, false)
 	}
+	if retErr != nil {
+		return nil, retErr
+	}
+
+	return vz.NewVirtioConsoleDeviceSerialPortConfiguration(serialPortAttachment)
+}
+
+func (dev *VirtioSerial) toVzConsole() (*vz.VirtioConsolePortConfiguration, error) {
+	master, slave, err := termios.Pty()
 	if err != nil {
 		return nil, err
 	}
 
-	return vz.NewVirtioConsoleDeviceSerialPortConfiguration(serialPortAttachment)
+	// the master fd and slave fd must stay open for vfkit's lifetime
+	util.RegisterExitHandler(func() {
+		_ = master.Close()
+		_ = slave.Close()
+	})
+
+	dev.PtyName = slave.Name()
+
+	if err := setRawMode(master); err != nil {
+		return nil, err
+	}
+	serialPortAttachment, retErr := vz.NewFileHandleSerialPortAttachment(master, master)
+	if retErr != nil {
+		return nil, retErr
+	}
+	return vz.NewVirtioConsolePortConfiguration(
+		vz.WithVirtioConsolePortConfigurationAttachment(serialPortAttachment),
+		vz.WithVirtioConsolePortConfigurationIsConsole(true))
 }
 
 func (dev *VirtioSerial) AddToVirtualMachineConfig(vmConfig *VirtualMachineConfiguration) error {
@@ -288,12 +322,24 @@ func (dev *VirtioSerial) AddToVirtualMachineConfig(vmConfig *VirtualMachineConfi
 	if dev.UsesStdio {
 		log.Infof("Adding stdio console")
 	}
+	if dev.PtyName != "" {
+		return fmt.Errorf("VirtioSerial.PtyName must be empty (current value: %s)", dev.PtyName)
+	}
 
-	consoleConfig, err := dev.toVz()
-	if err != nil {
-		return err
+	if dev.UsesPty {
+		consolePortConfig, err := dev.toVzConsole()
+		if err != nil {
+			return err
+		}
+		vmConfig.consolePortsConfiguration = append(vmConfig.consolePortsConfiguration, consolePortConfig)
+		log.Infof("Using PTY (pty path: %s)", dev.PtyName)
+	} else {
+		consoleConfig, err := dev.toVz()
+		if err != nil {
+			return err
+		}
+		vmConfig.serialPortsConfiguration = append(vmConfig.serialPortsConfiguration, consoleConfig)
 	}
-	vmConfig.serialPortsConfiguration = append(vmConfig.serialPortsConfiguration, consoleConfig)
 
 	return nil
 }
@@ -445,21 +491,62 @@ func AddToVirtualMachineConfig(vmConfig *VirtualMachineConfiguration, dev config
 	case *config.NetworkBlockDevice:
 		return (*NetworkBlockDevice)(d).AddToVirtualMachineConfig(vmConfig)
 	default:
-		return fmt.Errorf("Unexpected virtio device type: %T", d)
+		return fmt.Errorf("unexpected virtio device type: %T", d)
 	}
 }
 
-func (config *DiskStorageConfig) toVz() (vz.StorageDeviceAttachment, error) {
-	if config.ImagePath == "" {
-		return nil, fmt.Errorf("missing mandatory 'path' option for %s device", config.DevName)
+func (conf *DiskStorageConfig) toVz() (vz.StorageDeviceAttachment, error) {
+	switch conf.Type {
+	case config.DiskBackendImage, config.DiskBackendDefault:
+		if conf.ImagePath == "" {
+			return nil, fmt.Errorf("missing mandatory 'path' option for %s device", conf.DevName)
+		}
+		syncMode := vz.DiskImageSynchronizationModeFsync
+		caching := vz.DiskImageCachingModeCached
+		return vz.NewDiskImageStorageDeviceAttachmentWithCacheAndSync(conf.ImagePath, conf.ReadOnly, caching, syncMode)
+	case config.DiskBackendBlockDevice:
+		var stat unix.Stat_t
+		err := unix.Lstat(conf.ImagePath, &stat)
+
+		if err != nil {
+			return nil, fmt.Errorf("error stating file: %v", err)
+		}
+
+		mode := stat.Mode
+		if mode&unix.S_IFMT != unix.S_IFBLK {
+			return nil, fmt.Errorf("file %s is not a block device", conf.ImagePath)
+		}
+
+		var open_flags int
+		if conf.ReadOnly {
+			open_flags = os.O_RDONLY
+		} else {
+			open_flags = os.O_RDWR
+		}
+
+		f, err := os.OpenFile(conf.ImagePath, open_flags, 0)
+		if err != nil {
+			return nil, fmt.Errorf("error opening file: %v", err)
+		}
+
+		syncMode := vz.DiskSynchronizationModeFull
+		attachment, err := vz.NewDiskBlockDeviceStorageDeviceAttachment(f, conf.ReadOnly, syncMode)
+		if err != nil {
+			_ = f.Close()
+			return nil, fmt.Errorf("error creating disk attachment: %v", err)
+		}
+		util.RegisterExitHandler(func() {
+			f.Close()
+		})
+
+		return attachment, nil
+	default:
+		return nil, fmt.Errorf("unknown disk backend type: %v", conf.Type)
 	}
-	syncMode := vz.DiskImageSynchronizationModeFsync
-	caching := vz.DiskImageCachingModeCached
-	return vz.NewDiskImageStorageDeviceAttachmentWithCacheAndSync(config.ImagePath, config.ReadOnly, caching, syncMode)
 }
 
 func (dev *USBMassStorage) toVz() (vz.StorageDeviceConfiguration, error) {
-	var storageConfig DiskStorageConfig = DiskStorageConfig(dev.DiskStorageConfig)
+	var storageConfig = DiskStorageConfig(dev.DiskStorageConfig)
 	attachment, err := storageConfig.toVz()
 	if err != nil {
 		return nil, err
